/* ͻ *
 ͼ                File-Patch- und Dump-Utility                      ͻ
            Syntax: FILPATCH [D] <filename> [>stdout]                  
                                                                       
  => Mit der Option 'D' wird ein Filedump von <filename> erstellt.     
     Die Ausgabe kann mit [>stdout] umgeleitet werden.                 
                                                                       
  => Ohne Option gelangt man in einen Full-Screen-Editor, in dem das   
     File <filename> editiert werden kann. Dabei kann mit TAB zwischen 
     dem Hex- und dem ASCII-Fenster umgeschaltet werden. Mit den       
     Cursortasten kann der Cursor innerhalb des jeweiligen Fensters    
     positioniert werden. Zu ndernde Bytes/Zeichen werden einfach     
     berschrieben.                                                    
     Mit F1 kann direkt eine neue Position eingegeben werden; PgUp     
     und PgDn dagegen verndert die Position immer nur um 256 Bytes.   
     Mit ESC wird das Programm beendet.                                
     Wird die Position verndert oder das Programm beendet, wenn bis   
     dahin der aktuelle Fensterinhalt editiert wurde, erfolgt eine     
     Abfrage, ob der genderte Bereich abgespeichert werden soll.      
     Mit ESC kann man dies verneinen, mit CR besttigen.               
     Mit F6 besteht die Mglichkeit, das Programm kurzzeitig zu ver-   
     lassen, um auf die DOS-Ebene zu gelangen.                         
     Mit F5 kann ein neues File geladen werden. F2 und F3 sind fr     
     die Find-/Repeat-Find-Funktion zustndig. Es knnen ASCII-        
     Strings oder Hex-Bytefolgen gesucht werden, je nachdem, in        
 ͻ  welchem Fenster der Patch-Cursor gerade steht.                  ͼ
 * ͼ */

/*************************************************************************/
/* Headerfiles                                                           */
/*************************************************************************/
#include <stdio.h>
#include <dos.h>
#include <io.h>
#include <ioproc.h>       /* IOPROC.H: Schnelle Bildschirmroutinen! */
#include <bios.h>
#include <ctype.h>

/*************************************************************************/
/* Die Codes der Tasten                                                  */
/*************************************************************************/
#define CULE 75
#define CURI 77
#define CUUP 72
#define CUDN 80
#define PGUP 73
#define PGDN 81
#define F1 59
#define F2 60
#define F3 61
#define F4 62
#define F5 63
#define F6 64
#define TAB 9
#define CR 13
#define ESC 27

/*************************************************************************/
/* Globale Variablen, Flags und Puffer                                   */
/*************************************************************************/
unsigned char file_buffer[256],
              find_buffer[80],
              oldfile[50],
              curfile[16],
              patch_mode = 'h',      /* 'h' = HexPatch, 'a' = ASCIIPatch */
              *error;

char          patched = false,           /*  true: akt.Fenster gepatcht  */
              esc_pressed = false,
              found = false,
              ro = ' ';

unsigned int  read_bytes, taste,         /*  Anz. gelesene Bytes, ...    */
              find_len, i;

unsigned long file_pointer = 0, file_length;

FILE *fp;

/*************************************************************************/
/* Routine LIST fr Dump-Modus                                           */
/*************************************************************************/
void list (void)
{
    int l, b;
    unsigned char zeichen;

    printf ("OFFSET  00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F  0123456789ABCDEF\n\n");
    for (l=0; l<16; l++)
    {
         printf ("%.6X  ", file_pointer + l * 0x10);
         for (b=0; b<16; b++) printf ("%.2X ",file_buffer[l*0x10+b]);
         printf (" ");
         for (b=0; b<16; b++)
         {
              zeichen = file_buffer[l*0x10+b];
              if (zeichen < ' ' || zeichen == 0xFF)
              printf ("."); else printf ("%c", zeichen);
         }
         printf ("\n");
    }
    printf ("\n");
}

/*************************************************************************/
/* Routine DUMP fr Patch-Modus                                          */
/*************************************************************************/
void dump (void)
{
    int l, b, y, y1;
    unsigned long file_ofs;

    put_cursor (1,0);
    put_string ("  OFFSET    00 01 02 03 04 05 06 07 08 09 0A 0B 0C 0D 0E 0F   0123456789ABCDEF ");
    put_cursor (1,1);
    put_string ("          ͻͻ");

    for (l=0; l<16; l++)
    {
         put_cursor (1,l+2);
         put_string ("  ");
         put_hex (file_pointer + l * 0x10, 6);
         put_string ("   ");
         for (b=0; b<16; b++)
         {
              put_hex (file_buffer[l*0x10+b], 2);
              put_string (" ");
         }
         put_string ("");
         for (b=0; b<16; b++)
              put_char (file_buffer[l*0x10+b]);
         put_string ("");
    }
    put_cursor (1,18);
    put_string ("          ͹͹");
    put_cursor (1,19);
    put_string ("   Size:   F1-Offset  F2-Find  F3-Repeat-Find   PgUp  PgDn      File:      ");
    put_cursor (1,20);
    put_string ("  ");
    put_hex (file_length,6);
    put_string         ("   F5-Load  F6-DOS-Shell   TAB-Hex/ASCII  ESC-Quit  ");
    put_char (ro); put_string (curfile);
    put_string                                                                              ("  ");
    put_cursor (1,21);
    put_string ("          ͼͼ");

    get_cursor ();
    if (patch_mode == 'h')    /* Hex-Modus */
         file_ofs=file_pointer+(cursor_x-12)/3+16*(cursor_y-2);
    else                     /* ASCII-Modus */
         file_ofs=file_pointer+cursor_x-63+16*(cursor_y-2);

    y1=file_ofs*22/file_length;
    for (y=0; y<22; y++)
         put_cursor (0,y),
         (y1!=y) ? put_char ('') : (file_ofs==0) ? put_char (25) :
         (file_ofs>=file_length-1) ? put_char (24) : put_char (18);
}

/*************************************************************************/
/* Initialisiert den Bildschirm                                          */
/*************************************************************************/
void init_screen (void)
{
    video_karte ();
    save_screen ();
    vid_attrib = (vid_segment == 0xB000 ? 0x70 : 0x1F);
    cls ();
    clear_lines (22,24,7);
    set_cursor (13,2);
    full_cursor ();
}

/*************************************************************************/
/* Ermittelt den Namen der aktuellen Datei (ohne Pfadangabe).            */
/*************************************************************************/
void get_curfile (void)
{
    unsigned char a, b=0, i, l;

    l=strlen (oldfile); i=l;
    do i--; while (oldfile[i] != ':' && oldfile[i] != '\\' && i>0);
    if (oldfile[i] == ':' || oldfile[i] == '\\') i++;
    for (a=i; a<=l; a++, b++) curfile[b]=toupper (oldfile[a]);
    while (strlen (curfile)<12) strcat (curfile," ");
}

/*************************************************************************/
/* Druckt ein Zeichen an die aktuelle Cursorposition, ohne diese zu      */
/* inkrementieren.                                                       */
/*************************************************************************/
void print_char (unsigned char chr)
{
    AL = chr;
    AH = 9;
    BH = 0;
    BL = vid_attrib;
    CX = 1;
    INTBIOS
}

/*************************************************************************/
/* Fragen, ob Programmende (wenn ja, dann esc_flag = true)               */
/*************************************************************************/
void ende (void)
{
    get_cursor ();
    norm_cursor ();
    set_cursor (0,22);
    clear_lines (22,24,7);
    printf ("Do you want to quit (CR=Yes/ESC=No) ? ");

    do taste = getch ();
    while (taste != CR && taste != ESC);

    clear_lines (22,24,7);
    if (taste == CR) esc_pressed = true;
    set_cursor (cursor_x, cursor_y);
    full_cursor ();
}

/*************************************************************************/
/* Prft anhand des Flags 'patched', ob das aktuelle Fenster gapatcht    */
/* wurde und fragt ggf., ob es gesichert werden soll.                    */
/*************************************************************************/
void check_patch (void)
{
    if (patched && ro == ' ')
    {
         get_cursor ();
         norm_cursor ();
         set_cursor (0,22);
         clear_lines (22,24,7);
         printf ("Save patched Sector (CR=Yes/ESC=No) ? ");

         do taste = getch ();
         while (taste != CR && taste != ESC);

         clear_lines (22,24,7);
         if (taste == CR)
         {
              fseek (fp, ftell (fp)-read_bytes, SEEK_SET);
              if (fwrite (file_buffer, 1, read_bytes, fp) < read_bytes || ferror (fp))
              {
                   clearerr (fp);
                   set_cursor (0,22);
                   printf ("Error writing to File.");
              }
         }
         set_cursor (cursor_x, cursor_y);
         full_cursor ();
    }
    patched = false;
}

/*************************************************************************/
/* Liest ab 'offset' 256 Bytes in den File-Buffer ein und fllt ggf. den */
/* Rest mit 0 auf. Vorher wird evtl. der Pufferinhalt gesichert.         */
/*************************************************************************/
void read_buffer (unsigned long offset)
{
    int i;

    check_patch ();
    fseek (fp, offset, SEEK_SET);
    if ((read_bytes = fread (file_buffer, 1, 256, fp)) < 256)
    {
         for (i=read_bytes; i<256; i++) file_buffer[i] = 0;
    }
}

/*************************************************************************/
/* Wartet auf die Eingabe eines neuen Offsets, der angezeigt werden soll.*/
/*************************************************************************/
void position (void)
{
    unsigned long value;
    get_cursor ();
    clear_lines (22,24,7);
    norm_cursor ();
    set_cursor (0,22);
    printf ("Enter Offset/100H: ");
    while ((taste = bioskey (1)) == 0);
    taste = bioskey (1) & 0xFF;

    if (taste == CR || taste == ESC)
    {
         bioskey (0);
    }
    else
    {
         scanf ("%lx",&value);
         if (value * 0x100 < file_length)
         {
              file_pointer = value * 0x100;
              read_buffer (file_pointer);
         }
    }
    clear_lines (22,24,7);
    set_cursor (cursor_x, cursor_y);
    full_cursor ();
}

/*************************************************************************/
/* PgUp/PgDn, Cursortasten                                               */
/*************************************************************************/
void page_up (void)
{
    if (file_pointer > 0)
         file_pointer -= 0x100, read_buffer (file_pointer);
}

void page_dn (void)
{
    if (file_pointer+0x100 < file_length)
         file_pointer += 0x100, read_buffer (file_pointer);
}

void cursor_up (void)
{
    char y;
    get_cursor ();
    if ((y=cursor_y-1) < 2)
    {
         if (file_pointer > 0)
              y=17, page_up ();
         else y=cursor_y;
    }
    set_cursor (cursor_x, y);
}

void cursor_dn (void)
{
    char y;
    get_cursor ();
    if ((y=cursor_y+1) > 17)
    {
         if (file_pointer+0x100 < file_length)
              y = 2, page_dn ();
         else y=cursor_y;
    }
    set_cursor (cursor_x, y);
}

void h_cursor_left (void)
{
    char x,y;
    get_cursor ();
    x=cursor_x-1, y=cursor_y;
    if (x % 3 == 0) x--;
    if (x < 13)
    {
         x = 59; y--;
         if (y < 2)
         {
              if (file_pointer > 0)
                   y=17, page_up ();
              else y=cursor_y, x=cursor_x;
         }
    }
    set_cursor (x, y);
}

void h_cursor_right (void)
{
    char x,y;
    get_cursor ();
    x=cursor_x+1, y=cursor_y;
    if (x % 3 == 0) x++;
    if (x > 59)
    {
         x = 13; y++;
         if (y > 17)
         {
              if (file_pointer+0x100 < file_length)
                   y = 2, page_dn ();
              else y=cursor_y, x=cursor_x;
         }
    }
    set_cursor (x, y);
}

void a_cursor_left (void)
{
    char x,y;
    get_cursor ();
    x=cursor_x-1, y=cursor_y;
    if (x < 63)
    {
         x = 78; y--;
         if (y < 2)
         {
              if (file_pointer > 0)
                   y=17, page_up ();
              else y=cursor_y, x=cursor_x;
         }
    }
    set_cursor (x, y);
}

void a_cursor_right (void)
{
    char x,y;
    get_cursor ();
    x=cursor_x+1, y=cursor_y;
    if (x > 78)
    {
         x = 63; y++;
         if (y > 17)
         {
              if (file_pointer+0x100 < file_length)
                   y=2, page_dn ();
              else y=cursor_y, x=cursor_x;
         }
    }
    set_cursor (x, y);
}

/*************************************************************************/
/* Schaltet um zwischen Hex- und ASCII-Fenster.                          */
/*************************************************************************/
void switch_patch (void)
{
    if (patch_mode == 'h')
    {
         patch_mode = 'a';
         get_cursor ();
         cursor_x = (cursor_x-13)/3 + 63;
         set_cursor (cursor_x, cursor_y);
    }
    else
    {
         patch_mode = 'h';
         get_cursor ();
         cursor_x = (cursor_x-63)*3 + 13;
         set_cursor (cursor_x, cursor_y);
    }
}

/*************************************************************************/
/* Ldt COMMAND.COM oder sonstige Shell, die in 'COMSPEC' enthalten ist. */
/*************************************************************************/
void dos_shell (void)
{
    get_cursor ();
    vid_attrib = 7;
    cls ();
    norm_cursor ();
    set_cursor (0,0);
    printf ("Enter 'EXIT' to return to FilPatch.\n");
    system (getenv ("COMSPEC"));
    vid_attrib = (vid_segment == 0xB000 ? 0x70 : 0x1F);
    cls ();
    clear_lines (22,24,7);
    set_cursor (cursor_x, cursor_y);
    full_cursor ();
}

/*************************************************************************/
/* Patch-Taste gedrckt: Wert aus Bildschirmspeicher lesen und in Puffer */
/* in_buffer schreiben; patched-Flag = true setzen.                      */
/*************************************************************************/
void patch (int chr)
{
    unsigned char x_pos;
    unsigned int i;

    if (patch_mode == 'h')
    {
         chr = toupper (chr);
         if (((chr>='0') && (chr<='9')) || ((chr>='A') && (chr<='F')))
         {
              print_char (chr);
              get_cursor ();
              x_pos = (cursor_x-12)/3;
              i = cursor_y*160 + (x_pos*3+13)*2;
              in_buffer[0] = peekb (vid_segment, i);
              in_buffer[1] = peekb (vid_segment, i+2);
              in_buffer[2] = 0;
              file_buffer[x_pos + (cursor_y-2)*16] = strtol (in_buffer, error, 16);
              patched = true;
              h_cursor_right ();
         }
    }
    else
    {
         print_char (chr);
         get_cursor ();
         file_buffer[(cursor_y-2)*16 + cursor_x-63] = peekb (vid_segment, cursor_y*160 + cursor_x*2);
         patched = true;
         a_cursor_right ();
    }
}

/*************************************************************************/
/* Diese Routine durchsucht die Datei von Offset j ... file_length nach  */
/* den Zeichen in find_buffer, Anzahl = find_len.                        */
/*************************************************************************/
unsigned char scan (unsigned long j)
{
    unsigned long j1;
    unsigned char b=0, b1, i=0, byte[256];
    char bufchg;

    fseek (fp, j, SEEK_SET);
    fread (byte, 1, 256, fp);

    do
    {
         bufchg=false;
         if (byte[b]==find_buffer[i])
         {
              j1=j; b1=b;
              do
              {
                   i++; j++; b++;
                   if (b==0)
                   {
                        fseek (fp, j, SEEK_SET);
                        fread (byte, 1, 256, fp);
                        bufchg=true;
                   }
                   if (i==find_len)
                   {
                        found=true;
                        file_pointer= (j1>>8<<8);
                        return (j1-file_pointer);
                   }
                   if (j==file_length) return (0);
              }
              while (byte[b]==find_buffer[i]);
              i=0; j=j1; b=b1;
         }
         j++; b++;
         if (b==0 || bufchg)
         {
              fseek (fp, j, SEEK_SET);
              fread (byte, 1, 256, fp);
         }
    }
    while (j<file_length);
    return (0);
}

/*************************************************************************/
/* Suchen Hex-Bytefolgen oder ASCII-Strings, je nachdem, in welchem der  */
/* beiden Fenster sich der Patch-Cursor befindet.                        */
/*************************************************************************/
void find (void)
{
    unsigned char b, new_x, new_y;
    unsigned long j;

    check_patch ();
    found=false;
    clear_lines (22,24,7);
    norm_cursor ();
    set_cursor (0,22);
    printf ("Searching...");

    if (patch_mode == 'h')                                  /* Hex-Modus */
    {
         j=file_pointer+(cursor_x-12)/3+16*(cursor_y-2)+1;
         b=scan (j);
         new_x=(b%16)*3+13; new_y=b/16+2;
    }
    else                                                  /* ASCII-Modus */
    {
         j=file_pointer+cursor_x-63+16*(cursor_y-2)+1;
         b=scan (j);
         new_x=b%16+63; new_y=b/16+2;
    }

    read_buffer (file_pointer);
    set_cursor (0,22);

    if (found) printf ("Found.      "), set_cursor (new_x, new_y);
    else       printf ("Not found.  "), set_cursor (cursor_x, cursor_y);
    full_cursor ();
}

/*************************************************************************/
/* Eingabe der zu suchenden Hex-Bytefolge bzw. des ASCII-Strings, je     */
/* nachdem, in welchem der beiden Fenster sich der Patch-Cursor befindet */
/*************************************************************************/
void find_input (void)
{
    unsigned int i, j=0, k=0, l, value;

    clear_lines (22,24,7);
    norm_cursor ();
    set_cursor (0,22);

    if (patch_mode == 'h')
    {
    /* Hex-Modus: Eingabe der durch Kommas getrennten Hex-Zahlen in den  */
    /* Puffer in_buffer. Konvertierung in Zahlenwerte und bertragung    */
    /* nach find_buffer.                                                 */

         printf ("Enter Hex Byte Value(s) to find; separate them by commas:\n");
         while ((taste = bioskey (1)) == 0);
         taste = bioskey (1) & 0xFF;

         if (taste == CR || taste == ESC)
         {
              bioskey (0);
              set_cursor (cursor_x, cursor_y);
              clear_lines (22,24,7);
              full_cursor ();
         }
         else
         {
              scanf ("%s", &in_buffer);
              l=strlen (in_buffer);
              for (i=0; i<=l; i++)
              {
                   if (in_buffer[i] == ',' || in_buffer[i] == 0)
                   {
                        in_buffer[i] = 0;
                        value=strtol (&in_buffer[k], error, 16);
                        find_buffer[j]=value;
                        j++;
                        k=i+1;
                   }
              }
              find_len = j;
              find ();
         }
    }
    else
    /* ASCII-Modus: Direkte Eingabe des zu suchenden ASCII-Strings in    */
    /* den Puffer find_buffer.                                           */
    {
         printf ("Enter ASCII String to find:\n");
         while ((taste = bioskey (1)) == 0);
         taste = bioskey (1) & 0xFF;

         if (taste == CR || taste == ESC)
         {
              bioskey (0);
              set_cursor (cursor_x, cursor_y);
              clear_lines (22,24,7);
              full_cursor ();
         }
         else
         {
              scanf ("%s", &find_buffer);
              find_len = strlen (find_buffer);
              find ();
         }
    }
    /* Der Puffer find_buffer enthlt nun die zu suchende Bytefolge.    */
    /* Die Variable find_len enthlt die Anzahl der zu suchenden Bytes. */
}

/*************************************************************************/
/* Neue Datei in den Editor laden                                        */
/*************************************************************************/
void load_file (void)
{
    unsigned char filename[50];
    unsigned int x, y;

    get_cursor (), x = cursor_x, y = cursor_y;
    clear_lines (22,24,7);
    norm_cursor ();
    set_cursor (0,22);
    printf ("Enter new Filename: ");
    while ((taste = bioskey (1)) == 0);
    taste = bioskey (1) & 0xFF;

    if (taste == CR || taste == ESC)
    {
         bioskey (0);
         set_cursor (cursor_x, cursor_y);
    }
    else
    {
         scanf ("%s",&filename);
         check_patch ();
         fclose (fp);
         if ((fp = fopen (filename,"r+b")) == NULL)
         {
              if ((fp = fopen (filename,"rb")) == NULL)
              {
                   set_cursor (0,22);
                   printf ("Error opening File '%s'\n",filename);
                   if ((fp = fopen (oldfile,"r+b")) == NULL)
                   {
                        fp = fopen (oldfile,"rb"), ro = '*';
                   } else ro = ' ';
                   set_cursor (x, y);
                   full_cursor ();
                   return;
              } else ro = '*';
         } else ro = ' ';
         strcpy (oldfile, filename);
         get_curfile ();
         file_pointer=0;
         read_buffer (file_pointer);
         file_length = filelength (fileno (fp));
         if (patch_mode=='h') set_cursor (13,2);
         else                 set_cursor (63,2);
    }
    clear_lines (22,24,7);
    full_cursor ();
}

/*************************************************************************/
/*  H A U P T P R O G R A M M                                            */
/*************************************************************************/
void main (int argc, char *argv[])
{
    if (argc < 2 || argc > 3)
    {
         printf ("FILPATCH V2.00  (C) 1990-1991 Stefan Bion\n\n");
         printf ("Invocation Syntax: FILPATCH [D] <filename>\n");
         printf ("Option D -> Filedump to Stdout; No Option -> Edit File\n\n");
         printf ("Enter File to Edit: ");

         while ((taste = bioskey (1)) == 0);
         taste = bioskey (1) & 0xFF;
         if (taste == CR || taste == ESC) bioskey (0), exit (0);
         else scanf ("%s", oldfile); argc=2;
    }
    else strcpy (oldfile, argv[1]);

    if (argc == 2)

/*************************************************************************/
/*  a) Keine Option 'D': Patch-Modus.                                    */
/*************************************************************************/
    {
         if ((fp = fopen (oldfile,"r+b")) == NULL)
         {
              fp = fopen (oldfile,"rb"), ro = '*';
         } else ro = ' ';

         if (fp != NULL)
         {
              get_curfile ();
              file_length = filelength (fileno (fp));
              init_screen ();
              read_buffer (file_pointer);

              while (!esc_pressed)
              {
                   dump ();
                   taste = getch ();
                   switch (taste)
                   {
                        case 0:
                             switch (getch ())
                             {
                                  case F1:   position (); break;
                                  case F2:   get_cursor (); find_input (); break;
                                  case F3:   get_cursor (); find (); break;
                                  case F5:   load_file (); break;
                                  case F6:   dos_shell (); break;

                                  case PGUP: page_up (); break;
                                  case PGDN: page_dn (); break;

                                  case CUUP: cursor_up (); break;
                                  case CUDN: cursor_dn (); break;
                                  case CULE: (patch_mode == 'h') ? h_cursor_left () : a_cursor_left (); break;
                                  case CURI: (patch_mode == 'h') ? h_cursor_right () : a_cursor_right (); break;

                                  default:       break;
                             } break;

                        case TAB: switch_patch (); break;
                        case ESC: ende (); break;
                        default:  patch (taste & 0xFF); break;
                   }
              }
              check_patch ();
              norm_cursor ();
              restore_screen ();
              fclose (fp);
         }
         else printf ("Error opening File '%s'\n",oldfile);
    }
    else

/*************************************************************************/
/*  b) Option 'D': Dump-Modus.                                           */
/*************************************************************************/
    {
         if (*argv[1] == 'd' || *argv[1] == 'D')
         {
              if ((fp = fopen (argv[2],"rb")) != NULL)
              {
                   printf ("File: %s    Size: %lXh Bytes\n\n", argv[2],
                        filelength (fileno (fp)));
                   while (!feof (fp))
                   {
                        if ((read_bytes = fread (file_buffer, 1, 256, fp)) > 0)
                        {
                             if (read_bytes < 256)
                             {
                                  for (i=read_bytes; i<256; i++)
                                       file_buffer[i] = 0;
                             }
                             list ();
                             file_pointer += 0x100;
                        }
                   }
                   fclose (fp);
              }
              else printf ("Error opening File '%s'\n",argv[2]);
         }
         else printf ("Wrong Option.\n");
    }
}

/*************************************************************************/
/* Ende Ende Ende Ende Ende Ende Ende Ende Ende Ende Ende Ende Ende Ende */
/*************************************************************************/
